@@ -1,4 +1,4 @@
-/* $Id: acl.c,v 1.89 2009/02/09 04:12:07 manu Exp $ */
+/* $Id: acl.c,v 1.94 2009/09/07 12:56:54 manu Exp $ */
 
 /*
  * Copyright (c) 2004-2007 Emmanuel Dreyfus
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: acl.c,v 1.89 2009/02/09 04:12:07 manu Exp $");
+__RCSID("$Id: acl.c,v 1.94 2009/09/07 12:56:54 manu Exp $");
 #endif
 #endif
 
@@ -113,6 +113,7 @@ char *acl_print_regex(acl_data_t *, char *, size_t);
 char *acl_print_list(acl_data_t *, char *, size_t);
 char *acl_print_null(acl_data_t *, char *, size_t);
 char *acl_print_opnum(acl_data_t *, char *, size_t);
+char *acl_print_time(acl_data_t *, char *, size_t);
 int acl_opnum_cmp(int, enum operator, int);
 void acl_free_entry(struct acl_entry *);
 void acl_free_netblock(acl_data_t *);
@@ -126,6 +127,7 @@ void acl_add_body_regex(acl_data_t *, void *);
 void acl_add_macro(acl_data_t *, void *);
 void acl_add_opnum(acl_data_t *, void *);
 void acl_add_opnum_body(acl_data_t *, void *);
+void acl_add_time(acl_data_t *, void *);
 void acl_add_list(acl_data_t *, void *);
 char *acl_print_macro(acl_data_t *, char *, size_t);
 #ifdef USE_DNSRBL
@@ -378,6 +380,12 @@ struct acl_clause_rec acl_clause_rec[] = {
 	  AT_OPNUM, AC_NONE, AC_NONE,  EXF_SA,
 	  acl_print_opnum, acl_add_opnum, NULL, spamd_score },
 #endif
+#ifdef HAVE_DATA_CALLBACK
+	{ AC_TARPIT, UNIQUE, AS_ANY, "tarpit",
+	  AT_TIME, AC_NONE, AC_NONE, EXF_TARPIT,
+	  acl_print_time, acl_add_time,
+	  NULL, acl_tarpit_filter },
+#endif
 };
 
 struct {
@@ -548,6 +556,22 @@ acl_body_strstr(ad, stage, ap, priv)
 	return 0;
 }
 
+int
+acl_tarpit_filter(ad, stage, ap, priv)
+	acl_data_t *ad;
+	acl_stage_t stage;
+	struct acl_param *ap;
+	struct mlfi_priv *priv;
+{
+	ap->ap_tarpitted = pending_tarpitted(SA(&priv->priv_addr),
+					     priv->priv_addrlen,
+					     priv->priv_from,
+					     priv->priv_cur_rcpt);
+	ap->ap_tarpit = ad->time;
+	return ap->ap_tarpitted == (time_t)-1 || ap->ap_tarpitted > 0;
+}
+
+
 int
 myregexec(priv, ad, ap, string)
 	struct mlfi_priv *priv;
@@ -1006,6 +1030,16 @@ acl_print_opnum(ad, buf, len)
 	return buf;
 }
 
+char *
+acl_print_time(ad, buf, len)
+	acl_data_t *ad;
+	char *buf;
+	size_t len;
+{
+	snprintf(buf, len, "%ld", (long)ad->time);
+	return buf;
+}
+
 void
 acl_add_string(ad, data)
 	acl_data_t *ad;
@@ -1105,6 +1139,16 @@ acl_add_regex(ad, data)
 	return;
 }
 
+void
+acl_add_time(ad, data)
+	acl_data_t *ad;
+	void *data;
+{
+	time_t *t = (time_t *)data;
+	ad->time = *t;
+	return;
+}
+
 void
 acl_free_string(ad)
 	acl_data_t *ad;
@@ -1135,6 +1179,8 @@ acl_init_entry(void)
 	memset(acl, 0, sizeof(*acl));
 	acl->a_delay = -1;
 	acl->a_autowhite = -1;
+	acl->a_tarpit = -1;
+	acl->a_tarpit_scope = -1;
 	TAILQ_INIT(&acl->a_clause);
 
 	return acl;
@@ -1423,7 +1469,13 @@ acl_add_prop_string(ad, data)
 		exit(EX_OSERR);
 	}
 
-	acl_add_string((void *)&ad->prop->upd_data, upd->upd_data);
+	if ((ad->prop->upd_data = malloc(sizeof(acl_data_t))) == NULL) {
+		mg_log(LOG_ERR, "acl malloc failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+	memset(ad->prop->upd_data, 0, sizeof(acl_data_t));
+
+	acl_add_string((void *)ad->prop->upd_data, upd->upd_data);
 
 	return;
 }
@@ -1446,8 +1498,14 @@ acl_add_prop_regex(ad, data)
 		exit(EX_OSERR);
 	}
 
-	acl_add_regex((void *)&ad->prop->upd_data, upd->upd_data);
+	if ((ad->prop->upd_data = malloc(sizeof(acl_data_t))) == NULL) {
+		mg_log(LOG_ERR, "acl malloc failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
+	memset(ad->prop->upd_data, 0, sizeof(acl_data_t));
 
+	acl_add_regex((void *)ad->prop->upd_data, upd->upd_data);
+	
 	return;
 }
 
@@ -1460,8 +1518,7 @@ acl_print_prop_string(ad, buf, len)
 	size_t written;
 
 	written = snprintf(buf, len, "$%s ", ad->prop->upd_name);
-	acl_print_string((void *)&ad->prop->upd_data, 
-	    buf + written, len - written);
+	acl_print_string(ad->prop->upd_data, buf + written, len - written);
 
 	return buf;
 }
@@ -1475,8 +1532,7 @@ acl_print_prop_regex(ad, buf, len)
 	size_t written;
 
 	written = snprintf(buf, len, "$%s ", ad->prop->upd_name);
-	acl_print_regex((void *)&ad->prop->upd_data, 
-	    buf + written, len - written);
+	acl_print_regex(ad->prop->upd_data, buf + written, len - written);
 
 	return buf;
 }
@@ -1485,8 +1541,9 @@ void
 acl_free_prop_string(ad)
 	acl_data_t *ad;
 {
-	free(ad->prop->upd_name);
 	acl_free_string((void *)&ad->prop->upd_data);
+	free(ad->prop->upd_data);
+	free(ad->prop->upd_name);
 	free(ad->prop);
 
 	return;
@@ -1495,8 +1552,9 @@ void
 acl_free_prop_regex(ad)
 	acl_data_t *ad;
 {
-	free(ad->prop->upd_name);
 	acl_free_regex((void *)&ad->prop->upd_data);
+	free(ad->prop->upd_data);
+	free(ad->prop->upd_name);
 	free(ad->prop);
 
 	return;
@@ -1783,6 +1841,7 @@ acl_filter(stage, ctx, priv)
 	char addrstr[IPADDRSTRLEN];
 	char whystr[HDRLEN];
 	char tmpstr[HDRLEN];
+	char *aclstr;
 	int error = -1;
 	int retval = 0;
 	int noretval = 0;
@@ -1828,6 +1887,9 @@ acl_filter(stage, ctx, priv)
 		ap.ap_type = acl->a_type;
 		ap.ap_delay = acl->a_delay;
 		ap.ap_autowhite = acl->a_autowhite;
+		ap.ap_tarpit = acl->a_tarpit;
+		ap.ap_tarpitted = -1;
+		ap.ap_tarpit_scope = acl->a_tarpit_scope;
 		ap.ap_flags = acl->a_flags;
 		ap.ap_id = acl->a_id;
 		ap.ap_code = acl->a_code;
@@ -1909,6 +1971,15 @@ acl_filter(stage, ctx, priv)
 		priv->priv_sr.sr_autowhite =
 		    (ap.ap_autowhite != -1) ? 
 		    ap.ap_autowhite : conf.c_autowhite_validity;
+		priv->priv_sr.sr_tarpit =
+		    (ap.ap_tarpit != -1) ? ap.ap_tarpit : conf.c_tarpit;
+		priv->priv_sr.sr_tarpit_scope =
+		    (ap.ap_tarpit_scope != -1) ?
+		    ap.ap_tarpit_scope : conf.c_tarpit_scope;
+		if (ap.ap_tarpitted > priv->priv_max_tarpitted)
+			priv->priv_max_tarpitted = ap.ap_tarpitted;
+		if (ap.ap_tarpitted > priv->priv_total_tarpitted)
+			priv->priv_total_tarpitted = ap.ap_tarpitted;
 
 		if (ap.ap_id) {
 			priv->priv_sr.sr_acl_id = strdup(ap.ap_id);
@@ -2002,6 +2073,8 @@ acl_filter(stage, ctx, priv)
 
 		priv->priv_sr.sr_delay = conf.c_delay;
 		priv->priv_sr.sr_autowhite = conf.c_autowhite_validity;
+		priv->priv_sr.sr_tarpit = conf.c_tarpit;
+		priv->priv_sr.sr_tarpit_scope = conf.c_tarpit_scope;
 	}
 
 	if ((retval & EXF_NOLOG) == 0 && retval & EXF_WHITELIST) {
@@ -2072,14 +2145,22 @@ acl_filter(stage, ctx, priv)
 			ADD_REASON(whystr, tmpstr);
 		}		
 #endif
+		if (retval & EXF_TARPIT) {
+			snprintf(tmpstr, sizeof(tmpstr),
+			     "tarpit is%s requested",
+			    (noretval & EXF_TARPIT) ? notstr : vstr);
+			ADD_REASON(whystr, tmpstr);
+		}
 		if (retval & EXF_DEFAULT) {
 			ADD_REASON(whystr, "this is the default action");
 		}
 		iptostring(sa, salen, addrstr, sizeof(addrstr));
+		aclstr = fstring_expand(priv, NULL, "%a");
 		snprintf(tmpstr, sizeof(tmpstr),
-		    "(from=%s, rcpt=%s, addr=%s[%s]) ACL %d", from, 
+		    "(from=%s, rcpt=%s, addr=%s[%s]) ACL %s", from, 
 		    (cur_rcpt != NULL) ? cur_rcpt : "(nil)",
-		    hostname, addrstr, priv->priv_sr.sr_acl_line);
+		    hostname, addrstr, aclstr);
+		free(aclstr);
 		ADD_REASON(whystr, tmpstr);
 
 		mg_log(LOG_INFO, "%s: skipping greylist because %s",
@@ -2239,6 +2320,14 @@ acl_entry(entrystr, len, acl)
 		mystrlcat(entrystr, tempstr, len);
 	}
 
+	if (acl->a_tarpit_scope != -1) {
+		snprintf(tempstr, sizeof(tempstr),
+		    "[tarpit-scope \"%s\"] ",
+		    (acl->a_tarpit_scope == TAP_SESSION) ?
+		    "session" : "command");
+		mystrlcat(entrystr, tempstr, len);
+	}
+
 	if (acl->a_flags & A_FLUSHADDR) {
 		snprintf(tempstr, sizeof(tempstr), "[flushaddr] ");
 		mystrlcat(entrystr, tempstr, len);
@@ -2347,7 +2436,26 @@ acl_add_autowhite(delay)
 	gacl->a_autowhite = delay;
 		
 	if (conf.c_debug || conf.c_acldebug)
-		mg_log(LOG_DEBUG, "load acl delay %ld", (long)delay);
+		mg_log(LOG_DEBUG, "load acl autowhite %ld", (long)delay);
+
+	return;
+}
+
+void
+acl_add_tarpit_scope(scope)
+	tarpit_scope_t scope;
+{
+	if (gacl->a_tarpit_scope != -1) {
+		mg_log(LOG_ERR,
+		    "tarpit_scope specified twice in ACL line %d", conf_line);
+		exit(EX_DATAERR);
+	}
+
+	gacl->a_tarpit_scope = scope;
+
+	if (conf.c_debug || conf.c_acldebug)
+		mg_log(LOG_DEBUG, "load acl tarpit_scope %s",
+		   (scope == TAP_SESSION) ? "session" : "command");
 
 	return;
 }
@@ -2519,7 +2627,7 @@ acl_add_report(report)
 }
 
 void 
-acl_add_addheader(hdr)
+ acl_add_addheader(hdr)
 	char *hdr;
 {
 	if (gacl->a_addheader) {
@@ -2539,3 +2647,111 @@ acl_add_addheader(hdr)
 
 	return;
 }
+
+int
+acl_modify_by_prop(key, value, ap)
+	char *key;
+	char *value;
+	struct acl_param *ap;
+{
+	if (conf.c_acldebug)
+		mg_log(LOG_DEBUG, "check got \"%s\" => \"%s\"",
+		    key, value);
+	if (strcasecmp(key, "milterGreylistStatus") == 0) {
+		if ((strcasecmp(value, "Ok") == 0) ||
+		    (strcasecmp(value, "TRUE") == 0))
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistAction") == 0) {
+		if (strcasecmp(value, "greylist") == 0)
+			ap->ap_type = A_GREYLIST;
+		else if (strcasecmp(value, "blacklist") == 0)
+			ap->ap_type = A_BLACKLIST;
+		else if (strcasecmp(value, "whitelist") == 0)
+			ap->ap_type = A_WHITELIST;
+		else
+			mg_log(LOG_WARNING, "ignored greylist-type \"%s\"",
+			    value);
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistDelay") == 0) {
+		ap->ap_delay = humanized_atoi(value);
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistAutowhite") == 0) {
+		ap->ap_autowhite = humanized_atoi(value);
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistTarpit") == 0) {
+		ap->ap_tarpit = humanized_atoi(value);
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistFlushAddr") == 0) {
+		ap->ap_flags |= A_FLUSHADDR;
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistNoLog") == 0) {
+		ap->ap_flags |= A_NOLOG;
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistCode") == 0) {
+		if ((ap->ap_code = strdup(value)) == NULL) {
+			mg_log(LOG_ERR, "strdup(\"%s\") failed: %s",
+			    key, strerror(errno));
+			exit(EX_OSERR);
+		}
+		ap->ap_flags |= A_FREE_CODE;
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistEcode") == 0) {
+		if ((ap->ap_ecode = strdup(value)) == NULL) {
+			mg_log(LOG_ERR, "strdup(\"%s\") failed: %s",
+			    key, strerror(errno));
+			exit(EX_OSERR);
+		}
+		ap->ap_flags |= A_FREE_ECODE;
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistMsg") == 0) {
+		if ((ap->ap_msg = strdup(value)) == NULL) {
+			mg_log(LOG_ERR, "strdup(\"%s\") failed: %s",
+			    key, strerror(errno));
+			exit(EX_OSERR);
+		}
+		ap->ap_flags |= A_FREE_MSG;
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistReport") == 0) {
+		if ((ap->ap_report = strdup(value)) == NULL) {
+			mg_log(LOG_ERR, "strdup(\"%s\") failed: %s",
+			    key, strerror(errno));
+			exit(EX_OSERR);
+		}
+		ap->ap_flags |= A_FREE_REPORT;
+		goto out;
+	}
+
+	if (strcasecmp(key, "milterGreylistIgnore") == 0)
+		goto out;
+	if (conf.c_acldebug)
+		mg_log(LOG_DEBUG, "acl_modify inexpected property "
+		       "\"%s\"=\"%s\"", key, value);
+
+	return -1;
+out:
+	if (conf.c_acldebug)
+		mg_log(LOG_DEBUG, "modified acl property "
+		       "\"%s\"=\"%s\"", key, value);
+
+	return 0;
+}
