@@ -1,4 +1,4 @@
-/* $Id: conf.c,v 1.42.2.2 2006/09/20 07:38:24 manu Exp $ */
+/* $Id: conf.c,v 1.64 2009/09/09 12:19:17 manu Exp $ */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: conf.c,v 1.42.2.2 2006/09/20 07:38:24 manu Exp $");
+__RCSID("$Id: conf.c,v 1.64 2009/09/09 12:19:17 manu Exp $");
 #endif
 #endif
 
@@ -57,104 +57,184 @@ __RCSID("$Id: conf.c,v 1.42.2.2 2006/09/20 07:38:24 manu Exp $");
 #include <sys/stat.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <assert.h>
+#include <time.h>
+#include <sys/time.h>
 
+#include "spf.h"
 #include "acl.h"
 #ifdef USE_DNSRBL
 #include "dnsrbl.h"
 #endif
-#include "autowhite.h"
+#ifdef USE_CURL
+#include "urlcheck.h"
+#endif
+#ifdef USE_LDAP
+#include "ldapcheck.h"
+#endif
+#ifdef USE_DKIM
+#include "dkimcheck.h"
+#endif
+#ifdef USE_P0F
+#include "p0f.h"
+#endif
+#ifdef USE_SPAMD
+#include "spamd.h"
+#endif
 #include "conf.h"
 #include "sync.h"
 #include "pending.h"
 #include "dump.h"
+#include "list.h"
+#include "macro.h"
 #include "milter-greylist.h"
 
+#ifdef USE_DMALLOC
+#include <dmalloc.h> 
+#endif
+
+/* #define CONF_DEBUG */
+
 /* Default configuration */
-struct conf defconf;
-
-struct conf conf;
-
-int conf_cold = 1;
-int conf_nodetach = 0;
-
-char c_pidfile[QSTRLEN + 1];
-char c_dumpfile[QSTRLEN + 1];
-char c_socket[QSTRLEN + 1];
-char c_user[QSTRLEN + 1];
-char c_syncaddr[IPADDRSTRLEN + 1];
-char c_syncport[NUMLEN + 1];
-char c_syncsrcaddr[IPADDRSTRLEN + 1];
-char c_syncsrcport[NUMLEN + 1];
-char c_dracdb[QSTRLEN + 1];
-
+struct conf_rec defconf;
+pthread_key_t conf_key;
 char *conffile = CONFFILE;
-struct timeval conffile_modified;
-int numb_of_conf_update_threads;
+int conf_cold = 1;
+int conf_specified = 0;
 
-#define MAX_NUMB_OF_CONF_UPDATE_THREADS 1
-/*
- * this lock does not protect conf_update any more,
- * only conffile_modified and numb_of_conf_update_threads.
- * there are lot of non-auto variables above to protect as well,
- * so it is safer to limit the maximum number of configuration loading
- * processes to one for the time being.
- */
-pthread_rwlock_t conf_lock;
+#define CONF_LOCK pthread_mutex_lock(&conf_lock);
+#define CONF_UNLOCK pthread_mutex_unlock(&conf_lock);
+static pthread_mutex_t conf_lock = PTHREAD_MUTEX_INITIALIZER;
+static struct conf_list conf_list_head;
+static pthread_cond_t conf_update_cond = PTHREAD_COND_INITIALIZER;
+static int conf_updating;
+static char *legacy_conffile = "/etc/mail/greylist.conf";
 
 void
 conf_init(void) {
 	int error;
 
-	if ((error = pthread_rwlock_init(&conf_lock, NULL)) != 0) {
+	TAILQ_INIT(&conf_list_head);
+
+	if ((error = pthread_key_create(&conf_key, 0)) != 0) {
 		mg_log(LOG_ERR, 
-		    "pthread_rwlock_init failed: %s", strerror(error));
+		    "pthread_key_create failed: %s", strerror(error));
 		exit(EX_OSERR);
 	}
 
 	return;
 }
 
-void
-conf_load(void)
+#ifdef CONF_DEBUG
+static void
+conf_dump(void) {
+	struct conf_rec *c;
+
+	TAILQ_FOREACH_REVERSE(c, &conf_list_head, conf_list, c_chain) {
+		char textdate[DATELEN];
+		struct tm tm;
+
+		localtime_r(&c->c_timestamp, &tm);
+		strftime(textdate, sizeof textdate, "%Y-%m-%d %T",  &tm);
+		mg_log(LOG_DEBUG, "conf_dump: stamp %s ref %d",
+			   textdate, c->c_refcount);
+	}
+}
+#endif /* CONF_DEBUG */
+
+static void *
+conf_load_internal(timestamp)
+	void *timestamp;
 {
 	FILE *stream;
 	struct timeval tv1, tv2, tv3;
+	struct conf_rec *currconf, *threadconf, *newconf;
+
+	CONF_LOCK;
+	currconf = TAILQ_FIRST(&conf_list_head);
+	CONF_UNLOCK;
+	assert(conf_cold ? (currconf == NULL) : (currconf != NULL));
+	threadconf = GET_CONF();
+
+	if (!(newconf = (struct conf_rec *)malloc(sizeof *newconf))) {
+		mg_log(LOG_ERR, "conf malloc failed: %s", strerror(errno));
+		exit(EX_OSERR);
+	}
 
 	/*
 	 * Reset the configuration to its default 
 	 * (This includes command line flags)
 	 */
-	memcpy(&conf, &defconf, sizeof(conf));
+	memcpy(newconf, &defconf, sizeof *newconf);
+	newconf->c_refcount = 1;
+	newconf->c_timestamp = *(time_t *)timestamp;
 
 	(void)gettimeofday(&tv1, NULL);
 
-	if (!conf_cold || conf.c_debug)
+	if (!conf_cold || newconf->c_debug)
 		mg_log(LOG_INFO, "%sloading config file \"%s\"", 
 		    conf_cold ? "" : "re", conffile);
 
-	if ((stream = fopen(conffile, "r")) == NULL) {
+	if (!conf_specified && strcmp(conffile, legacy_conffile) != 0) {
+		struct stat st;
+
+		if (stat(legacy_conffile, &st) == 0) {
+			mg_log(LOG_WARNING,
+			       "legacy configuration file '%s' is used "
+			       "instead of '%s'. IT SHOULD BE RENAMED!",
+			       legacy_conffile,
+			       conffile);
+			conffile = legacy_conffile;
+		}
+	}
+
+	errno = 0;
+	if ((stream = Fopen(conffile, "r")) == NULL) {
 		mg_log(LOG_ERR, "cannot open config file %s: %s", 
-		    conffile, strerror(errno));
-		mg_log(LOG_ERR, "continuing with no exception list");
+		    conffile, 
+		    (errno == 0) ? "out of stdio streams" : strerror(errno));
+
+		if (conf_cold)
+			exit(EX_OSERR);
 	} else {
+		TSS_SET(conf_key, newconf);
 
 		peer_clear();
 		ACL_WRLOCK;
 #ifdef USE_DNSRBL
 		dnsrbl_clear();
 #endif
+#ifdef USE_CURL
+		urlcheck_clear();
+#endif
+#ifdef USE_LDAP
+		ldapcheck_clear();
+#endif
+#ifdef USE_DKIM
+		dkimcheck_clear();
+#endif
+#ifdef USE_P0F
+		p0f_clear();
+#endif
+		all_list_clear();
+		macro_clear();
 		acl_clear();
 
 		conf_in = stream;
 		conf_line = 1;
 		conf_acl_end = 0;
+		conf_racl_end = 0;
+		conf_dacl_end = 0;
 
 		conf_parse();
+		conf_dispose_input_file();
 		ACL_UNLOCK;
 
-		fclose(stream);
+		TSS_SET(conf_key, threadconf);
 
-		if (!conf_cold || conf.c_debug) {
+		Fclose(stream);
+
+		if (!conf_cold || newconf->c_debug) {
 			(void)gettimeofday(&tv2, NULL);
 			timersub(&tv2, &tv1, &tv3);
 			mg_log(LOG_INFO,
@@ -164,18 +244,36 @@ conf_load(void)
 		}
 	}
 
-	if (conf_cold) {
-		(void)gettimeofday(&conffile_modified, NULL);
-	} else {
-		CONF_WRLOCK;
-		--numb_of_conf_update_threads;
-		CONF_UNLOCK;
-	}
-
-	if (conf.c_debug || conf.c_acldebug)
+	/*
+	 * Dump the ACL for debugging purposes
+	 */
+	if (newconf->c_debug || newconf->c_acldebug)
 		acl_dump();
 
-	return;
+	CONF_LOCK;
+	assert(TAILQ_FIRST(&conf_list_head) == currconf);
+	if (currconf && --currconf->c_refcount == 0) {
+		TAILQ_REMOVE(&conf_list_head, currconf, c_chain);
+		free(currconf);
+	}
+	TAILQ_INSERT_HEAD(&conf_list_head, newconf, c_chain);
+	CONF_UNLOCK;
+
+#ifdef CONF_DEBUG
+	conf_dump();
+#endif
+	dump_conf_changed();
+	return NULL;
+}
+
+/* Functions other than main() must not invoke this */
+void
+conf_load(void) {
+	struct stat st;
+
+	if (stat(conffile, &st))
+		st.st_mtime = (time_t)0;
+	conf_load_internal(&st.st_mtime);
 }
 
 void
@@ -184,6 +282,7 @@ conf_update(void) {
 	pthread_t tid;
 	pthread_attr_t attr;
 	int error;
+	int need_update;
 	
 	if (stat(conffile, &st) != 0) {
 		mg_log(LOG_ERR, "config file \"%s\" unavailable", 
@@ -191,17 +290,16 @@ conf_update(void) {
 		return;
 	}
 
-	CONF_WRLOCK;
-	numb_of_conf_update_threads++;
-	if (st.st_mtime <= conffile_modified.tv_sec ||
-		numb_of_conf_update_threads > MAX_NUMB_OF_CONF_UPDATE_THREADS) {
-		--numb_of_conf_update_threads;
-		CONF_UNLOCK;
-		return;
-	}
-	conffile_modified.tv_sec = st.st_mtime;
+	CONF_LOCK;
+	while (conf_updating)
+		pthread_cond_wait(&conf_update_cond, &conf_lock);
+	conf_updating = need_update =
+		st.st_mtime > TAILQ_FIRST(&conf_list_head)->c_timestamp;
 	CONF_UNLOCK;
 
+	if (!need_update)
+		return;
+
 	/*
 	 * On some platforms, the thread stack limit is too low and
 	 * conf_parse will get a SIGSEGV because it overflows the
@@ -229,27 +327,97 @@ conf_update(void) {
 	}
 
 	if ((error = pthread_create(&tid, &attr, 
-	    (void *(*)(void *))conf_load, NULL)) != 0) {
+	    conf_load_internal, &st.st_mtime)) != 0) {
 		mg_log(LOG_ERR, "pthread_create failed: %s", 
 		    strerror(error));
 		exit(EX_OSERR);
 	}
 
-	if ((error = pthread_detach(tid)) != 0) {
-		mg_log(LOG_ERR, "pthread_detach failed: %s",
-		    strerror(error));
-		exit(EX_OSERR);
-	}
-
 	if ((error = pthread_attr_destroy(&attr)) != 0) {
 		mg_log(LOG_ERR, "pthread_attr_destroy failed: %s",
 		    strerror(error));
 		exit(EX_OSERR);
 	}
 
+	if ((error = pthread_join(tid, NULL)) != 0) {
+		mg_log(LOG_ERR, "pthread_join failed: %s",
+		    strerror(error));
+		exit(EX_OSERR);
+	}
+
+	CONF_LOCK;
+	conf_updating = 0;
+	CONF_UNLOCK;
+	if ((error = pthread_cond_broadcast(&conf_update_cond)) != 0) {
+		mg_log(LOG_ERR, "pthread_cond_broadcast failed: %s", 
+		       strerror(error));
+		abort();
+	}
+
 	return;
 }
 
+void
+conf_retain(void) {
+	struct conf_rec *c;
+
+	if (GET_CONF()) {
+		mg_log(LOG_ERR, "%s:%d BUG: conf_retain called twice?",
+				__FILE__, __LINE__);
+		assert(0);
+	}
+
+	CONF_LOCK;
+	c = TAILQ_FIRST(&conf_list_head);
+#ifdef CONF_DEBUG
+	{
+		char textdate[DATELEN];
+		struct tm tm;
+		
+		localtime_r(&c->c_timestamp, &tm);
+		strftime(textdate, sizeof textdate, "%Y-%m-%d %T",  &tm);
+		mg_log(LOG_DEBUG, "conf_retain: stamp %s ref %d -> %d",
+		       textdate, c->c_refcount, c->c_refcount + 1);
+	}
+#endif
+	++c->c_refcount;
+	CONF_UNLOCK;
+
+	TSS_SET(conf_key, c);
+}
+
+void
+conf_release(void) {
+	struct conf_rec *c = GET_CONF();
+
+	if (!c) {
+		mg_log(LOG_ERR, "%s:%d BUG: conf_release before conf_retain",
+				__FILE__, __LINE__);
+		assert(0);
+		return;
+	}
+
+	CONF_LOCK;
+#ifdef CONF_DEBUG
+	{
+		char textdate[DATELEN];
+		struct tm tm;
+		
+		localtime_r(&c->c_timestamp, &tm);
+		strftime(textdate, sizeof textdate, "%Y-%m-%d %T",  &tm);
+		mg_log(LOG_DEBUG, "conf_release: stamp %s ref %d -> %d",
+		       textdate, c->c_refcount, c->c_refcount - 1);
+	}
+#endif
+	if (--c->c_refcount == 0) {
+		TAILQ_REMOVE(&conf_list_head, c, c_chain);
+		free(c);
+	}
+	CONF_UNLOCK;
+
+	TSS_SET(conf_key, NULL);
+}
+
 /*
  * Write path into dst, stripping leading and trailing quotes
  */
@@ -272,8 +440,11 @@ quotepath(dst, path, len)
 
 void
 conf_defaults(c)
-	struct conf *c;
+	struct conf_rec *c;
 {
+	c->c_refcount = -1;
+	c->c_timestamp = (time_t)0;
+
 	c->c_forced = C_GLNONE;
 	c->c_debug = 0;
 	c->c_acldebug = 0;
@@ -285,8 +456,11 @@ conf_defaults(c)
 	c->c_testmode = 0;
 	c->c_delay = GLDELAY;
 	c->c_autowhite_validity = AUTOWHITE_VALIDITY;
+	c->c_tarpit = TARPIT_DURATION;
+	c->c_tarpit_scope = TAP_SESSION;
 	c->c_pidfile = NULL;
 	c->c_dumpfile = DUMPFILE;
+	c->c_dumpfile_mode = -1;
 	prefix2mask4(32, &c->c_match_mask);
 #ifdef AF_INET6
 	prefix2mask6(128, &c->c_match_mask6);
@@ -296,13 +470,25 @@ conf_defaults(c)
 	c->c_syncsrcaddr = NULL;
 	c->c_syncsrcport = NULL;
 	c->c_socket = NULL;
+	c->c_socket_mode = -1;
 	c->c_user = NULL;
+	c->c_nodetach = 0;
 	c->c_report = C_ALL;
 	c->c_dumpfreq = DUMPFREQ;
 	c->c_timeout = TIMEOUT;
 	c->c_extendedregex = 0;
 	c->c_dracdb = DRACDB;
 	c->c_nodrac = 0;
-
+	c->c_maxpeek = 0;
+	c->c_logfac = LOG_MAIL;
+	c->c_domainexact = 0;
+#ifdef USE_P0F
+	c->c_p0fsock[0] = '\0';
+#endif
+#ifdef USE_SPAMD
+	c->c_spamdsock[0] = '\0';
+	c->c_spamdsocktype[0] = '\0';
+#endif
+	c->c_syncmaxqlen = SYNC_MAXQLEN;
 	return;
 }
