@@ -1,4 +1,4 @@
-/* $Id: dump.c,v 1.35 2007/12/29 19:06:49 manu Exp $ */
+/* $Id: dump.c,v 1.40 2009/09/26 14:38:39 manu Exp $ */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID  
-__RCSID("$Id: dump.c,v 1.35 2007/12/29 19:06:49 manu Exp $");
+__RCSID("$Id: dump.c,v 1.40 2009/09/26 14:38:39 manu Exp $");
 #endif
 #endif
 
@@ -67,7 +67,6 @@ __RCSID("$Id: dump.c,v 1.35 2007/12/29 19:06:49 manu Exp $");
 #include "conf.h"
 #include "sync.h"
 #include "dump.h"
-#include "autowhite.h"
 #include "milter-greylist.h"
 
 #ifdef USE_DMALLOC
@@ -84,7 +83,7 @@ __RCSID("$Id: dump.c,v 1.35 2007/12/29 19:06:49 manu Exp $");
 
 static pthread_mutex_t dump_todo_lock = PTHREAD_MUTEX_INITIALIZER;
 static int dump_todo = 0;
-static pthread_cond_t dump_sleepflag;
+static pthread_cond_t dump_sleepflag = PTHREAD_COND_INITIALIZER;
 #define DUMP_TODO_CONF_UPDATE 0x1
 #define DUMP_TODO_FLUSH 0x2
 #define DUMP_TODO_TERMINATE 0x4
@@ -96,19 +95,6 @@ static int dump_dirty = 0;
 
 static pthread_t dumper_tid;
 
-void
-dump_init(void) {
-	int error;
-
-	if ((error = pthread_cond_init(&dump_sleepflag, NULL)) != 0) {
-		mg_log(LOG_ERR, 
-		    "pthread_cond_init failed: %s", strerror(error));
-		exit(EX_OSERR);
-	}
-
-	return;
-}
-
 void
 dumper_start(void) {
 	int error;
@@ -204,11 +190,10 @@ dump_perform(final)
 {
 	FILE *dump;
 	int dumpfd;
+	int done;
 	struct timeval tv1, tv2, tv3;
 	char newdumpfile[MAXPATHLEN + 1];
-	int done;
-	int greylisted_count;
-	int whitelisted_count;
+	tuple_cnt_t greylisted_count;
 	char *s_buffer = NULL;
 	int dirty;
 
@@ -281,13 +266,25 @@ dump_perform(final)
 	
 	dump_header(dump);
 	greylisted_count = pending_textdump(dump);
-	whitelisted_count = autowhite_textdump(dump);
-	done = greylisted_count + whitelisted_count;
 
-	fprintf(dump, "#\n# Summary: %d records, %d greylisted, %d "
-	    "whitelisted\n#\n", done, greylisted_count, whitelisted_count);
+	done = greylisted_count.pending + greylisted_count.autowhite + 
+	       greylisted_count.tarpit;
+
+	fprintf(dump, "#\n# Summary: %d records, %d greylisted, "
+		"%d whitelisted, %d tarpitted\n#\n", done,
+		greylisted_count.pending, 
+		greylisted_count.autowhite,
+		greylisted_count.tarpit);
+
+	/*
+	 * Ensure that the data is really flushed to disk.
+	 * Some systems might delay the data write from kernel buffer
+	 * to disk even after the file is closed
+	 */
+	(void)fflush(dump);
+	(void)fsync(fileno(dump));
+	(void)Fclose(dump);
 
-	Fclose(dump);
 	if (s_buffer)
 		free(s_buffer);
 
@@ -322,12 +319,10 @@ dump_reload(void) {
 	} else {
 		dump_in = dump;
 		PENDING_LOCK;
-		AUTOWHITE_LOCK;
 
 		dump_parse();
 		dump_dispose_input_file();
 
-		AUTOWHITE_UNLOCK;
 		PENDING_UNLOCK;
 		Fclose(dump);
 
