@@ -1,4 +1,4 @@
-/* $Id: milter-greylist.c,v 1.215 2008/11/26 05:20:13 manu Exp $ */
+/* $Id: milter-greylist.c,v 1.222 2009/09/09 12:19:17 manu Exp $ */
 
 /*
  * Copyright (c) 2004-2007 Emmanuel Dreyfus
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID  
-__RCSID("$Id: milter-greylist.c,v 1.215 2008/11/26 05:20:13 manu Exp $");
+__RCSID("$Id: milter-greylist.c,v 1.222 2009/09/09 12:19:17 manu Exp $");
 #endif
 #endif
 
@@ -86,7 +86,7 @@ static int check_drac(char *dotted_ip);
 #include "conf.h"
 #include "pending.h"
 #include "sync.h"
-#include "autowhite.h"
+#include "store.h"
 #include "stat.h"
 #include "milter-greylist.h"
 #ifdef USE_DNSRBL
@@ -124,6 +124,7 @@ static void smtp_reply_init(struct smtp_reply *);
 static void smtp_reply_free(struct smtp_reply *);
 static void add_recipient(struct mlfi_priv *, char *);
 static void set_sr_defaults(struct mlfi_priv *, char *, char *, char *);
+static sfsistat tarpit_reentry(struct mlfi_priv *);
 static sfsistat stat_from_code(char *);
 static void cleanup_pidfile(char *);
 static void cleanup_sock(char *);
@@ -136,7 +137,11 @@ static sfsistat real_header(SMFICTX *, char *, char *);
 static sfsistat real_eoh(SMFICTX *);
 static sfsistat real_body(SMFICTX *, unsigned char *, size_t);
 static sfsistat real_eom(SMFICTX *);
+static sfsistat real_abort(SMFICTX *);
 static sfsistat real_close(SMFICTX *);
+#ifdef HAVE_DATA_CALLBACK
+static sfsistat real_data(SMFICTX *);
+#endif
 
 struct smfiDesc smfilter =
 {
@@ -151,8 +156,12 @@ struct smfiDesc smfilter =
 	mlfi_eoh,	/* end of header */
 	mlfi_body,	/* body block filter */
 	mlfi_eom,	/* end of message */
-	NULL,		/* message aborted */
+	mlfi_abort,	/* message aborted */
 	mlfi_close,	/* connection cleanup */
+#ifdef HAVE_DATA_CALLBACK
+	NULL,		/* any unrecognized or unimplemented command filter */
+	mlfi_data,	/* SMTP DATA command filter */
+#endif
 };
 
 static int nodetach = 0;
@@ -266,6 +275,18 @@ mlfi_eom(ctx)
 	return r;
 }
 
+sfsistat
+mlfi_abort(ctx)
+	SMFICTX *ctx;
+{
+	sfsistat r;
+
+	conf_retain();
+	r = real_abort(ctx);
+	conf_release();
+	return r;
+}
+
 sfsistat
 mlfi_close(ctx)
 	SMFICTX *ctx;
@@ -278,6 +299,43 @@ mlfi_close(ctx)
 	return r;
 }
 
+#ifdef HAVE_DATA_CALLBACK
+sfsistat
+mlfi_data(ctx)
+	SMFICTX *ctx;
+{
+	sfsistat r;
+
+	conf_retain();
+	r = real_data(ctx);
+	conf_release();
+	return r;
+}
+#endif
+
+static sfsistat
+tarpit_reentry(priv)
+	struct mlfi_priv *priv;
+{
+	sfsistat stat = SMFIS_CONTINUE;
+
+	if (!(priv->priv_sr.sr_whitelist & EXF_TARPIT))
+		return stat;
+
+	if (priv->priv_sr.sr_whitelist & EXF_WHITELIST) {
+		struct rcpt *rcpt;
+
+		rcpt = priv->priv_rcpt.lh_first;
+		pending_update(SA(&priv->priv_addr), priv->priv_addrlen,
+			       priv->priv_from, rcpt->r_addr,
+			       priv->priv_sr.sr_autowhite, TU_AUTOWHITE);
+	} else if (priv->priv_sr.sr_whitelist & EXF_GREYLIST) {
+		stat = SMFIS_TEMPFAIL;
+	}
+
+	return stat;
+}
+
 static sfsistat
 real_connect(ctx, hostname, addr)
 	SMFICTX *ctx;
@@ -353,6 +411,9 @@ real_connect(ctx, hostname, addr)
 	priv->priv_p0f = NULL;
 	p0f_lookup(priv);
 #endif
+	priv->priv_max_tarpitted = 0;
+	priv->priv_total_tarpitted = 0;
+
 	return SMFIS_CONTINUE;
 }
 
@@ -524,7 +585,8 @@ real_envrcpt(ctx, envrcpt)
 	char addrstr[IPADDRSTRLEN];
 	char rcpt[ADDRLEN + 1];
 	int save_nolog;
-
+	struct tuple_fields tuple;
+	time_t sleep_duration = 0;
 	/*
 	 * Strip spaces from the recipient address
 	 */
@@ -535,6 +597,8 @@ real_envrcpt(ctx, envrcpt)
 		mg_log(LOG_ERR, "Internal error: smfi_getpriv() returns NULL");
 		return SMFIS_TEMPFAIL;
 	}
+	if (tarpit_reentry(priv) == SMFIS_TEMPFAIL)
+		return SMFIS_TEMPFAIL;
 
 	if (!iptostring(SA(&priv->priv_addr), priv->priv_addrlen, addrstr,
 	    sizeof(addrstr)))
@@ -602,7 +666,8 @@ real_envrcpt(ctx, envrcpt)
 		return SMFIS_TEMPFAIL;
 	}
 
-	if (priv->priv_sr.sr_whitelist & EXF_WHITELIST) {
+	if (priv->priv_sr.sr_whitelist & EXF_WHITELIST &&
+	    priv->priv_sr.sr_tarpit <= 0) {
 		priv->priv_sr.sr_elapsed = 0;
 		goto exit_accept;
 	}
@@ -611,14 +676,12 @@ real_envrcpt(ctx, envrcpt)
 	 * Blacklist overrides autowhitelisting...
 	 */
 	if (priv->priv_sr.sr_whitelist & EXF_BLACKLIST) {
-		char aclstr[16];
+		char *aclstr;
 		char *code = "551";
 		char *ecode = "5.7.1";
 		char *msg = "Go away!";
 
-		if (priv->priv_sr.sr_acl_line != 0)
-			snprintf(aclstr, sizeof(aclstr), " (ACL %d)", 
-			    priv->priv_sr.sr_acl_line);
+		aclstr = fstring_expand(priv, NULL, " (ACL %a)");
 
 		if (!(priv->priv_sr.sr_whitelist & EXF_NOLOG)) {
 			mg_log(LOG_INFO, 
@@ -629,37 +692,67 @@ real_envrcpt(ctx, envrcpt)
 
 		set_sr_defaults(priv, code, ecode, msg);
 		mg_setreply(ctx, priv, rcpt);
+		free(aclstr);
 		return mg_stat(priv, stat_from_code(priv->priv_sr.sr_code));
 	}
 
-	/* 
-	 * Check if the tuple {sender IP, sender e-mail, recipient e-mail}
-	 * was autowhitelisted.
-	 */
 	save_nolog = priv->priv_sr.sr_whitelist & EXF_NOLOG;
-	priv->priv_sr.sr_whitelist = autowhite_check(SA(&priv->priv_addr),
-	    priv->priv_addrlen, priv->priv_from, rcpt, priv->priv_queueid,
-	    priv->priv_sr.sr_delay, priv->priv_sr.sr_autowhite);
-
-	if (priv->priv_sr.sr_whitelist != EXF_NONE) {
-		priv->priv_sr.sr_elapsed = 0;
-		goto exit_accept;
-	}
 
 	/*
 	 * Check if the tuple {sender IP, sender e-mail, recipient e-mail}
-	 * is in the greylist and if it ca now be accepted. If it is not
-	 * in the greylist, it will be added.
+	 * is in the greylist or autowhite list and if it can now be 
+	 * accepted. If in greylist, the greylist entry will change to 
+	 * autowhite entry. If it is not in the greylist, it will be added.
 	 */
-	if (pending_check(SA(&priv->priv_addr), priv->priv_addrlen,
-	    priv->priv_from, rcpt, &remaining, &priv->priv_sr.sr_elapsed,
-	    priv->priv_queueid, priv->priv_sr.sr_delay, 
-	    priv->priv_sr.sr_autowhite) != 0) {
+
+	tuple.sa = SA(&priv->priv_addr);
+	tuple.salen = priv->priv_addrlen;
+	tuple.from = priv->priv_from;
+	tuple.rcpt = rcpt;
+	tuple.remaining = &remaining;
+	tuple.elapsed = &priv->priv_sr.sr_elapsed;
+	tuple.queueid = priv->priv_queueid;
+	tuple.gldelay = priv->priv_sr.sr_delay;
+	tuple.autowhite = priv->priv_sr.sr_autowhite;
+
+	switch(mg_tuple_check(tuple)) {
+	case T_AUTOWHITE:			/* autowhite listed */
+		priv->priv_sr.sr_elapsed = 0;
+		priv->priv_sr.sr_whitelist = EXF_WHITELIST | EXF_AUTO;
+		goto exit_accept;
+		break;
+	case T_PENDING:			/* greylisted */
 		if (priv->priv_sr.sr_elapsed > priv->priv_max_elapsed)
 			priv->priv_max_elapsed = priv->priv_sr.sr_elapsed;
 		goto exit_accept;
-	}
+		break;
+	default:			/* first encounter */
+		if (!(priv->priv_sr.sr_whitelist & EXF_TARPIT))
+			break;
 
+		if (priv->priv_sr.sr_tarpit_scope == TAP_COMMAND &&
+		    priv->priv_sr.sr_tarpit > priv->priv_max_tarpitted)
+			sleep_duration = priv->priv_sr.sr_tarpit;
+		if (priv->priv_sr.sr_tarpit_scope == TAP_SESSION &&
+		    priv->priv_sr.sr_tarpit > priv->priv_total_tarpitted)
+			sleep_duration = priv->priv_sr.sr_tarpit -
+					      priv->priv_total_tarpitted;
+
+		if (sleep_duration > 0) {
+			if (sleep_duration > priv->priv_max_tarpitted)
+				priv->priv_max_tarpitted = sleep_duration;
+			priv->priv_total_tarpitted += sleep_duration;
+			sleep(sleep_duration);
+		} else {
+		    if (priv->priv_sr.sr_whitelist & EXF_WHITELIST)
+			pending_update(SA(&priv->priv_addr), priv->priv_addrlen,
+				       priv->priv_from, rcpt,
+				       priv->priv_sr.sr_autowhite,
+				       TU_AUTOWHITE);
+		}
+		priv->priv_sr.sr_elapsed = 0;
+		goto exit_accept;
+	}
 	priv->priv_sr.sr_remaining = remaining;
 
 	/*
@@ -843,7 +936,7 @@ real_body(ctx, chunk, size)
 			priv->priv_buf = NULL;
 		}
 
-		memcpy(b->b_lines + priv->priv_buflen, chunk, i + 1);
+		memcpy(b->b_lines + priv->priv_buflen, chunk, i);
 		b->b_lines[linelen] = '\0';
 		priv->priv_buflen = 0;
 
@@ -877,9 +970,9 @@ real_eom(ctx)
 	struct smtp_reply rcpt_sr;
 	struct rcpt *rcpt;
 	time_t remaining;
-	int whitelist;
 	int envrcpt_continue = 0;
 	int accept = 1;
+	struct tuple_fields tuple;
 
 	if ((priv = (struct mlfi_priv *) smfi_getpriv(ctx)) == NULL) {
 		mg_log(LOG_ERR, "Internal error: smfi_getpriv() returns NULL");
@@ -939,16 +1032,15 @@ real_eom(ctx)
 	}
 
 	if (priv->priv_sr.sr_whitelist & EXF_BLACKLIST) {
-		char aclstr[16];
+		char *aclstr;
 		char addrstr[IPADDRSTRLEN];
 		char *code = "551";
 		char *ecode = "5.7.1";
 		char *msg = "Go away!";
 
 		smtp_reply_free(&rcpt_sr);
-		if (priv->priv_sr.sr_acl_line != 0)
-			snprintf(aclstr, sizeof(aclstr), " (ACL %d)", 
-			    priv->priv_sr.sr_acl_line);
+
+		aclstr = fstring_expand(priv, NULL, " (ACL %a)");
 
 		iptostring(SA(&priv->priv_addr), priv->priv_addrlen, addrstr,
 		    sizeof(addrstr));
@@ -962,44 +1054,48 @@ real_eom(ctx)
 
 		set_sr_defaults(priv, code, ecode, msg);
 		mg_setreply(ctx, priv, NULL);
+		free(aclstr);
 		return mg_stat(priv, stat_from_code(priv->priv_sr.sr_code));
 	}
 
 	if (priv->priv_sr.sr_whitelist & EXF_GREYLIST && envrcpt_continue) {
-		/* 
-	 	 * Check if the tuple {sender IP, sender e-mail, 
-		 * recipient e-mail} was autowhitelisted 
-	 	 */
+		/*
+		 * Multiple recipients for a single message. Here we check 
+		 * each recipient individually for greylisting and autowhite 
+		 * listing.
+		 */
 		priv->priv_sr.sr_whitelist &= ~EXF_NONE;
 		LIST_FOREACH(rcpt, &priv->priv_rcpt, r_list) {
-			whitelist = 
-			    autowhite_check(SA(&priv->priv_addr),
-	    				    priv->priv_addrlen, 
-					    priv->priv_from, 
-					    rcpt->r_addr, 
-					    priv->priv_queueid,
-	    				    priv->priv_sr.sr_delay,
-					    priv->priv_sr.sr_autowhite);
-			priv->priv_sr.sr_whitelist |= whitelist;
 
-			if ((whitelist == EXF_NONE) && 
-			    !pending_check(SA(&priv->priv_addr), 
-					   priv->priv_addrlen,
-	    				   priv->priv_from, 
-					   rcpt->r_addr, 
-					   &remaining, 
-					   &priv->priv_sr.sr_elapsed,
-	    				   priv->priv_queueid, 
-					   priv->priv_sr.sr_delay, 
-	    				   priv->priv_sr.sr_autowhite))
+			tuple.sa = SA(&priv->priv_addr);
+			tuple.salen = priv->priv_addrlen;
+			tuple.from = priv->priv_from;
+			tuple.rcpt = rcpt->r_addr;
+			tuple.remaining = &remaining;
+			tuple.elapsed = &priv->priv_sr.sr_elapsed;
+			tuple.queueid = priv->priv_queueid;
+			tuple.gldelay = priv->priv_sr.sr_delay;
+			tuple.autowhite = priv->priv_sr.sr_autowhite;
+
+			switch(mg_tuple_check(tuple)) {
+			case T_AUTOWHITE:	/* autowhite listed */
+				priv->priv_sr.sr_whitelist |= 
+				    (EXF_WHITELIST | EXF_AUTO);
+				break;
+			case T_PENDING:		/* greylisted */
+				break;
+			default:		/* first encounter */
 				accept = 0;
+				break;
+			}
+
 		}
 
 		if (accept) {
 			if (priv->priv_sr.sr_elapsed > priv->priv_max_elapsed)
 				priv->priv_max_elapsed = 
 				    priv->priv_sr.sr_elapsed;
-				goto passed;
+			goto passed;
 		}
 
 		priv->priv_sr.sr_remaining = remaining;
@@ -1123,13 +1219,18 @@ passed:
 				    "Default is to whitelist mail");
 				priv->priv_last_whitelist &= ~EXF_DEFAULT;
 			}
+			if (priv->priv_last_whitelist & EXF_TARPIT) {
+				ADD_REASON(whystr,
+				    "Message whitelisted by tarpit %ts");
+				priv->priv_last_whitelist &= ~EXF_TARPIT;
+			}
 			priv->priv_last_whitelist &= 
 			    ~(EXF_GREYLIST | EXF_WHITELIST | EXF_BLACKLIST);
 
 			/* Silently ignore other codes, just report ACL */	
 			if (priv->priv_last_whitelist != 0) {
 				priv->priv_last_whitelist = 0;
-				mystrlcat (whystr, "ACL %A matched", HDRLEN);
+				mystrlcat (whystr, "ACL %a matched", HDRLEN);
 			}
 
 			mystrlcat (whystr, ", not delayed by %V", HDRLEN);
@@ -1179,6 +1280,27 @@ out:
 	return mg_stat(priv, SMFIS_CONTINUE);
 }
 
+static sfsistat
+real_abort(ctx)
+	SMFICTX *ctx;
+{
+	struct mlfi_priv *priv;
+
+	if ((priv = (struct mlfi_priv *) smfi_getpriv(ctx)) != NULL) {
+		if (priv->priv_sr.sr_whitelist & EXF_TARPIT) {
+			struct rcpt *rcpt;
+
+			rcpt = priv->priv_rcpt.lh_first;
+			pending_update(SA(&priv->priv_addr), priv->priv_addrlen,
+				       priv->priv_from, rcpt->r_addr,
+				       0, TU_TARPIT);
+			priv->priv_sr.sr_whitelist &= ~EXF_TARPIT;
+		}
+	}
+
+	return SMFIS_CONTINUE;
+}
+
 static sfsistat
 real_close(ctx)
 	SMFICTX *ctx;
@@ -1232,7 +1354,21 @@ real_close(ctx)
 	return SMFIS_CONTINUE;
 }
 
+#ifdef HAVE_DATA_CALLBACK
+static sfsistat
+real_data(ctx)
+	SMFICTX *ctx;
+{
+	struct mlfi_priv *priv;
 
+	if ((priv = (struct mlfi_priv *) smfi_getpriv(ctx)) == NULL) {
+		mg_log(LOG_ERR, "Internal error: smfi_getpriv() returns NULL");
+		return SMFIS_TEMPFAIL;
+	}
+
+	return tarpit_reentry(priv);
+}
+#endif
 
 int
 main(argc, argv)
@@ -1269,6 +1405,7 @@ main(argc, argv)
 			    (time_t)humanized_atoi(optarg);
 			defconf.c_forced |= C_AUTOWHITE;
 			break;
+
 		case 'c':
 		        checkonly = 1;
 			break;
@@ -1335,6 +1472,7 @@ main(argc, argv)
 				usage(argv[0]);
 			}
 			conffile = optarg;
+			conf_specified = 1;
 			break;
 
 		case 'd':
@@ -1451,10 +1589,8 @@ main(argc, argv)
 	conf_init();
 	all_list_init();
 	acl_init ();
-	pending_init();
 	peer_init();
-	autowhite_init();
-	dump_init();
+
 #ifdef USE_DNSRBL
 	dnsrbl_init();
 #endif
@@ -1522,10 +1658,9 @@ main(argc, argv)
 	(void)smfi_setconn(conf.c_socket);
 
 	/*
-	 * Reload a saved greylist
-	 * No lock needed here either.
+	 * Initialize the storage backend
 	 */
-	dump_reload();
+	mg_init();
 
 	/*
 	 * If no body/header search exists, don't install the hooks,
@@ -1664,15 +1799,9 @@ main(argc, argv)
 	pthread_sigmask(SIG_BLOCK, &set, NULL);
 
 	/*
-	 * Start the dumper thread
+	 * Start the storage backend background thread
 	 */
-	dumper_start();
-
-	/*
-	 * Run the peer MX greylist sync threads
-	 */
-	sync_master_restart();
-	sync_sender_start();
+	mg_start();
 
 	/*
 	 * Here we go!
@@ -1689,8 +1818,9 @@ main(argc, argv)
 	dump_perform(1);
 	conf_release();
 #else
-	dumper_stop();
+	mg_tuple_stop();	/* stop storage background threads */
 #endif
+	mg_tuple_close();	/* close storage backend */
 	return exitval;
 }
 
@@ -1959,7 +2089,7 @@ log_and_report_greylisting(ctx, priv, rcpt)
 	char addrstr[IPADDRSTRLEN];
 	time_t remaining;
 	char *delayed_rj;
-	char aclstr[16];
+	char *aclstr;
 	char *code = "451";
 	char *ecode = "4.7.1";
 	char *msg = conf.c_quiet ?
@@ -1985,11 +2115,7 @@ log_and_report_greylisting(ctx, priv, rcpt)
 	else
 		delayed_rj = "";
 
-	if (priv->priv_sr.sr_acl_line != 0)
-		snprintf(aclstr, sizeof(aclstr), " (ACL %d)", 
-		    priv->priv_sr.sr_acl_line);
-	else
-		aclstr[0] = '\0';
+	aclstr = fstring_expand(priv, NULL, " (ACL %a)");
 
 	if (!(priv->priv_sr.sr_whitelist & EXF_NOLOG)) {
 		mg_log(LOG_INFO, 
@@ -2000,6 +2126,7 @@ log_and_report_greylisting(ctx, priv, rcpt)
 
 	set_sr_defaults(priv, code, ecode, msg);
 	mg_setreply(ctx, priv, rcpt);
+	free(aclstr);
 	return;
 }
 
@@ -2057,6 +2184,7 @@ smtp_reply_init(sr)
 	/* sr->sr_elapsed = (time_t)0xdeadbeefU; */
 	sr->sr_delay = conf.c_delay;
 	sr->sr_autowhite = conf.c_autowhite_validity;
+	sr->sr_tarpit = conf.c_tarpit;
 
 	return;
 }
@@ -2949,6 +3077,14 @@ fstring_expand(priv, rcpt, fstring)
 			}
 			break;
 		}	
+		case 't': {	/* tarpit time */
+			char buf[32];
+
+			snprintf(buf, sizeof(buf),
+				 "%ld", (long)priv->priv_sr.sr_tarpit);
+			mystrncat(&outstr, buf, &outmaxlen);
+			break;
+		}
 		case '%':	/* Literal '%' */
 			mystrncat(&outstr, "%", &outmaxlen);
 			break;
