@@ -1,4 +1,5 @@
-/* $Id: pending.h,v 1.37.2.1 2006/09/04 22:05:59 manu Exp $ */
+/* $Id: pending.h,v 1.40 2009/09/07 12:56:54 manu Exp $ */
+/* vim: set sw=8 ts=8 sts=8 noet cino=(0: */
 
 /*
  * Copyright (c) 2004 Emmanuel Dreyfus
@@ -45,10 +46,18 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
+#ifndef AUTOWHITE_VALIDITY
+#define AUTOWHITE_VALIDITY (24 * 3600) /* 1 day */
+#endif
+
 #ifndef GLDELAY
 #define GLDELAY	1800	/* 1800 seconds = 30 minutes */
 #endif
 
+#ifndef TARPIT_DURATION
+#define TARPIT_DURATION -1 /* do not tarpit */
+#endif
+
 #ifndef TIMEOUT
 #define TIMEOUT (3600 * 24 * 5) /* 432000 seconds = 5 days */
 #endif
@@ -59,9 +68,17 @@
 
 #include "milter-greylist.h"
 
-#define PENDING_WRLOCK WRLOCK(pending_lock)
-#define PENDING_RDLOCK RDLOCK(pending_lock)
-#define PENDING_UNLOCK UNLOCK(pending_lock)
+typedef enum {
+	T_TARPIT = 3,
+	T_AUTOWHITE = 2,
+	T_PENDING = 1,
+	T_NONE = 0
+} tuple_t;
+
+typedef enum {
+	TU_AUTOWHITE,
+	TU_TARPIT
+} tuple_update_type_t;
 
 TAILQ_HEAD(pendinglist, pending);
 
@@ -71,28 +88,43 @@ struct pending {
 	socklen_t p_salen;
 	char *p_from;
 	char *p_rcpt;
-	struct timeval p_tv;
-	int p_refcnt;
+	struct timeval p_tv;	/* activation time for pending or tarpitting, 
+				   expiration time for autowhite */
+	struct timeval p_tarpit;/* unsuccessful longest duration for tarpit */
+	int p_refcnt;		/* mutex; pending entry is 
+				   shared with sync queue */
+        tuple_t	p_type;		/* pending or autowhite entry? */
 	TAILQ_ENTRY(pending) p_list;
 	TAILQ_ENTRY(pending) pb_list;
 };
 
 struct pending_bucket {
-	pthread_mutex_t	bucket_mtx;
 	TAILQ_HEAD(, pending) b_pending_head;
 };
 
-extern pthread_rwlock_t pending_lock;
-extern pthread_mutex_t pending_change_lock;
+typedef struct {
+	int pending;
+	int autowhite;
+	int tarpit;
+} tuple_cnt_t;
+
+#define PENDING_LOCK pthread_mutex_lock(&pending_lock);
+#define PENDING_UNLOCK pthread_mutex_unlock(&pending_lock);
+extern pthread_mutex_t pending_lock;
 
 void pending_init(void);
 struct pending *pending_get(struct sockaddr *, socklen_t, char *, char *,
+    time_t, tuple_t);
+tuple_t pending_check(struct sockaddr *, socklen_t, char *, char *, 
+    time_t *, time_t *, char *, time_t, time_t);
+time_t pending_tarpitted(struct sockaddr *, socklen_t, char *, char *);
+void pending_update(struct sockaddr *, socklen_t, char *, char *, 
+    time_t, tuple_update_type_t);
+void pending_del(struct sockaddr *, socklen_t, char *, char *, time_t, 
     time_t);
-int pending_check(struct sockaddr *, socklen_t, char *, char *, time_t *,
-    time_t *, char *, time_t, time_t);
-void pending_del(struct sockaddr *, socklen_t, char *, char *, time_t);
-void pending_put(struct pending *);
-int pending_textdump(FILE *);
+void pending_rem(struct pending *);
+void pending_put(struct pending *, time_t);
+tuple_cnt_t pending_textdump(FILE *);
 struct pending *pending_ref(struct pending *);
 void pending_free(struct pending *);
 int ip_match(struct sockaddr *, struct sockaddr *, ipaddr *);
