@@ -1,7 +1,7 @@
-/* $Id: sync.c,v 1.66.2.5 2006/11/07 05:12:12 manu Exp $ */
+/* $Id: sync.c,v 1.86 2009/05/23 22:50:53 manu Exp $ */
 
 /*
- * Copyright (c) 2004 Emmanuel Dreyfus
+ * Copyright (c) 2004-2007 Emmanuel Dreyfus
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,7 +34,7 @@
 #ifdef HAVE_SYS_CDEFS_H
 #include <sys/cdefs.h>
 #ifdef __RCSID
-__RCSID("$Id: sync.c,v 1.66.2.5 2006/11/07 05:12:12 manu Exp $");
+__RCSID("$Id: sync.c,v 1.86 2009/05/23 22:50:53 manu Exp $");
 #endif
 #endif
 
@@ -56,9 +56,12 @@ __RCSID("$Id: sync.c,v 1.66.2.5 2006/11/07 05:12:12 manu Exp $");
 #include "pending.h"
 #include "sync.h"
 #include "conf.h"
-#include "autowhite.h"
 #include "milter-greylist.h"
 
+#ifdef USE_DMALLOC
+#include <dmalloc.h> 
+#endif
+
 #define SYNC_PROTO_CURRENT 3
 
 struct sync_master_sock {
@@ -66,12 +69,49 @@ struct sync_master_sock {
 	int sock;
 };
 
+#define SMS_RUNNING		1
+#define SMS_NOTRUNNING		0
+#define SMS_DISABLED		-1
+
+/* errors returned by stdio that should not cause an exit */
+static int sync_ignored_errno[] = { 
+#ifdef EAGAIN
+	EAGAIN,
+#endif /* EAGAIN */
+#ifdef ECONNABORTED
+	ECONNABORTED,
+#endif /* ECONNABORTED */
+#ifdef EMFILE
+	EMFILE,
+#endif /* EMFILE */
+#ifdef ENFILE
+	ENFILE,
+#endif /* ENFILE */
+#ifdef ENOBUFS
+	ENOBUFS,
+#endif /* ENOBUFS */
+#ifdef ENOMEM
+	ENOMEM,
+#endif /* ENOMEM */
+#ifdef ENOSR
+	ENOSR,
+#endif /* ENOSR */
+#ifdef EWOULDBLOCK
+	EWOULDBLOCK,
+#endif /* EWOULDBLOCK */
+	0,
+};
+
+static pthread_mutex_t sync_master_lock = PTHREAD_MUTEX_INITIALIZER;
 struct sync_master_sock sync_master4 = { 0, -1 };
 struct sync_master_sock sync_master6 = { 0, -1 };
+
 struct peerlist peer_head;
 pthread_rwlock_t peer_lock; /* For the peer list */
-pthread_rwlock_t sync_lock; /* For all peer's sync queue */
-pthread_cond_t sync_sleepflag;
+
+static pthread_mutex_t sync_dirty_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_cond_t sync_sleepflag = PTHREAD_COND_INITIALIZER;
+static int sync_dirty = 0;
 
 static void sync_listen(char *, char *, struct sync_master_sock *);
 static int local_addr(const struct sockaddr *sa, const socklen_t salen);
@@ -79,6 +119,7 @@ static int sync_queue_poke(struct peer *, struct sync *);
 static struct sync * sync_queue_peek(struct peer *);
 static int select_protocol(struct peer *, int, FILE *);
 static void sync_vers(FILE *, int);
+static int is_fatal(int);
 
 void
 peer_init(void) {
@@ -91,18 +132,6 @@ peer_init(void) {
 		exit(EX_OSERR);
 	}
 
-	if ((error = pthread_rwlock_init(&sync_lock, NULL)) != 0) {
-		mg_log(LOG_ERR, 
-		    "pthread_rwlock_init failed: %s", strerror(error));
-		exit(EX_OSERR);
-	}
-
-	if ((error = pthread_cond_init(&sync_sleepflag, NULL)) != 0) {
-		mg_log(LOG_ERR, 
-		    "pthread_cond_init failed: %s", strerror(error));
-		exit(EX_OSERR);
-	}
-
 	return;
 }
 
@@ -120,9 +149,10 @@ peer_clear(void) {
 			sync_free(sync);
 			
 		if (peer->p_stream != NULL)
-			fclose(peer->p_stream);
+			Fclose(peer->p_stream);
 
 		LIST_REMOVE(peer, p_list);
+		pthread_mutex_destroy(&peer->p_mtx);
 		free(peer->p_name);
 		free(peer);
 	}
@@ -133,8 +163,9 @@ peer_clear(void) {
 }
 
 void 
-peer_add(peername)
+peer_add(peername, sockettimeout)
 	char *peername;
+        time_t sockettimeout;
 {
 	struct peer *peer;
 
@@ -147,7 +178,9 @@ peer_add(peername)
 	peer->p_qlen = 0;
 	peer->p_stream = NULL;
 	peer->p_flags = 0;
+	peer->p_socket_timeout = sockettimeout;
 	TAILQ_INIT(&peer->p_deferred);
+	pthread_mutex_init(&peer->p_mtx, NULL);
 
 	PEER_WRLOCK;
 	LIST_INSERT_HEAD(&peer_head, peer, p_list);
@@ -203,16 +236,16 @@ sync_queue_poke(peer, sync)
 	struct peer *peer;
 	struct sync *sync;
 {
-	SYNC_WRLOCK;
-	if (peer->p_qlen < SYNC_MAXQLEN) {
+	int r = 0;
+
+	pthread_mutex_lock(&peer->p_mtx);
+	if (peer->p_qlen < conf.c_syncmaxqlen) {
 		TAILQ_INSERT_HEAD(&peer->p_deferred, sync, s_list);
 		peer->p_qlen++;
-		SYNC_UNLOCK;
-		return 1;
-	} else {
-		SYNC_UNLOCK;
-		return 0;
+		r = 1;
 	}
+	pthread_mutex_unlock(&peer->p_mtx);
+	return r;
 }
 
 static struct sync *
@@ -221,13 +254,13 @@ sync_queue_peek(peer)
 {
 	struct sync *sync;
 
-	SYNC_WRLOCK;
+	pthread_mutex_lock(&peer->p_mtx);
 	sync = TAILQ_FIRST(&peer->p_deferred);
 	if (!TAILQ_EMPTY(&peer->p_deferred)) {
 		TAILQ_REMOVE(&peer->p_deferred, sync, s_list);
 		peer->p_qlen--;
 	}
-	SYNC_UNLOCK;
+	pthread_mutex_unlock(&peer->p_mtx);
 	return sync;
 }
 
@@ -246,6 +279,7 @@ sync_send(peer, type, pending, autowhite) /* peer list is read-locked */
 	char *keyw;
 	char awstr[LINELEN + 1];
 	int bw;
+	int errcount = 0;
 
 	if ((peer->p_stream == NULL) && (peer_connect(peer) != 0))
 		return -1;
@@ -278,22 +312,22 @@ sync_send(peer, type, pending, autowhite) /* peer list is read-locked */
 	}
 
 	if (bw > LINELEN) {
-		mg_log(LOG_ERR, "closing connexion with peer %s: "
+		mg_log(LOG_ERR, "closing connection with peer %s: "
 		    "send buffer would overflow (%d entries queued)", 
 		    peer->p_name, peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
 
 	bw = fprintf(peer->p_stream, "%s", line);
 	if (bw != strlen(line)) {
-		mg_log(LOG_ERR, "closing connexion with peer %s: "
+		mg_log(LOG_ERR, "closing connection with peer %s: "
 		    "%s (%d entries queued) - I was unable to send "
 		    "complete line \"%s\" - bytes written: %i", 
 		    peer->p_name, strerror(errno), peer->p_qlen, 
 		    line, bw);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
@@ -304,23 +338,28 @@ sync_send(peer, type, pending, autowhite) /* peer list is read-locked */
 	 * Check the return code 
 	 */
 	get_more:
-	sync_waitdata(peer->p_socket);
+	sync_waitdata(peer->p_socket, peer->p_socket_timeout);
 	if (fgets(line, LINELEN, peer->p_stream) == NULL) {
 		if (errno == EAGAIN) {
+			errcount++;
 			if ( feof(peer->p_stream) ) {
-				mg_log(LOG_ERR, "lost connexion with peer %s: "
+				mg_log(LOG_ERR, "lost connection with peer %s: "
 		  		  "%s (%d entries queued)", 
 				    peer->p_name, strerror(errno), peer->p_qlen);
-				fclose(peer->p_stream);
+				Fclose(peer->p_stream);
 				peer->p_stream = NULL;
 				return -1;
 			}
-			goto get_more;
+			if (errcount < 5) {
+				goto get_more;
+			} else {
+				mg_log(LOG_ERR, "too many read error with peer %s", peer->p_name);
+			}
 		}
-		mg_log(LOG_ERR, "lost connexion with peer %s: "
+		mg_log(LOG_ERR, "lost connection with peer %s: "
 		    "%s (%d entries queued)", 
 		    peer->p_name, strerror(errno), peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
@@ -334,9 +373,9 @@ sync_send(peer, type, pending, autowhite) /* peer list is read-locked */
 
 	if ((replystr = strtok_r(line, sep, &cookie)) == NULL) {
 		mg_log(LOG_ERR, "Unexpected reply \"%s\" from %s, "
-		    "closing connexion (%d entries queued)", 
+		    "closing connection (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
@@ -344,9 +383,9 @@ sync_send(peer, type, pending, autowhite) /* peer list is read-locked */
 	replycode = atoi(replystr);
 	if (replycode != 201) {
 		mg_log(LOG_ERR, "Unexpected reply \"%s\" from %s, "
-		    "closing connexion (%d entries queued)", 
+		    "closing connection (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
-		fclose(peer->p_stream);
+		Fclose(peer->p_stream);
 		peer->p_stream = NULL;
 		return -1;
 	}
@@ -539,21 +578,28 @@ peer_connect(peer)	/* peer list is read-locked */
 		    peer->p_name, strerror(errno));
 	}
 
-	if ((stream = fdopen(s, "w+")) == NULL) {
+	errno = 0;
+	if ((stream = Fdopen(s, "w+")) == NULL) {
 		mg_log(LOG_ERR, "cannot sync with peer %s, "
 		    "fdopen failed: %s (%d entries queued)", 
-		    peer->p_name, strerror(errno), peer->p_qlen);
+		    peer->p_name, 
+		    (errno == 0) ? "out of stdio streams" : strerror(errno),
+		    peer->p_qlen);
 		close(s);
 		return -1;
 	}
 
+#ifdef USE_FD_POOL
+	s = fileno(stream);     /* the socket descriptor could have been replaced by Fdopen() ! */
+#endif
+
 	if (setvbuf(stream, NULL, _IOLBF, 0) != 0)
 		mg_log(LOG_ERR, "cannot set line buffering with peer %s: %s", 
 		    peer->p_name, strerror(errno));
 
-	sync_waitdata(s);	
+	sync_waitdata(s, peer->p_socket_timeout);	
 	if (fgets(line, LINELEN, stream) == NULL) {
-		mg_log(LOG_ERR, "Lost connexion with peer %s: "
+		mg_log(LOG_ERR, "Lost connection with peer %s: "
 		    "%s (%d entries queued)", 
 		    peer->p_name, strerror(errno), peer->p_qlen);
 		goto bad;
@@ -568,7 +614,7 @@ peer_connect(peer)	/* peer list is read-locked */
 
 	if ((replystr = strtok_r(line, sep, &cookie)) == NULL) {
 		mg_log(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
-		    "closing connexion (%d entries queued)", 
+		    "closing connection (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
 		goto bad;
 	}
@@ -576,7 +622,7 @@ peer_connect(peer)	/* peer list is read-locked */
 	replycode = atoi(replystr);
 	if (replycode != 200) {
 		mg_log(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
-		    "closing connexion (%d entries queued)", 
+		    "closing connection (%d entries queued)", 
 		    line, peer->p_name, peer->p_qlen);
 		goto bad;
 	}
@@ -592,49 +638,80 @@ peer_connect(peer)	/* peer list is read-locked */
 	return 0;
 
 bad:
-	fclose(stream);
+	Fclose(stream);
 	peer->p_stream = NULL;
 
 	return -1;
 }
 
+void
+sync_master_stop(void) {
+	if (sync_master4.sock != 1) {
+		close(sync_master4.sock);
+		sync_master4.sock = -1;
+	}
+
+	if (sync_master6.sock != 1) {
+		close(sync_master6.sock);
+		sync_master6.sock = -1;
+	}
+}
+
 void
 sync_master_restart(void) {
 	pthread_t tid;
 	int empty;
 	int error;
+	int sync_master6_runs = SMS_NOTRUNNING;
+	int sync_master4_runs = SMS_NOTRUNNING;
 
 	PEER_RDLOCK;
 	empty = LIST_EMPTY(&peer_head);
 	PEER_UNLOCK;
 
-	if (empty || sync_master4.runs || sync_master6.runs)
-		return;
+	pthread_mutex_lock(&sync_master_lock);
+	if (empty || (sync_master4.runs && sync_master6.runs))
+		goto last;
 
 	if (conf.c_syncaddr != NULL) {
-		if (strchr(conf.c_syncaddr, ':'))
-		    sync_listen(conf.c_syncaddr, conf.c_syncport,
-				&sync_master6);
-		else
-		    sync_listen(conf.c_syncaddr, conf.c_syncport,
-				&sync_master4);
+		if (strchr(conf.c_syncaddr, ':')) {
+			if (sync_master6.runs != SMS_RUNNING) {
+				sync_listen(conf.c_syncaddr, conf.c_syncport,
+					    &sync_master6);
+				sync_master6_runs = sync_master6.runs;
+			}
+			sync_master4.runs = SMS_DISABLED;
+		} else {
+			if (sync_master4.runs != SMS_RUNNING) {
+				sync_listen(conf.c_syncaddr, conf.c_syncport,
+					    &sync_master4);
+				sync_master4_runs = sync_master4.runs;
+			}
+			sync_master6.runs = SMS_DISABLED;
+		}
 	} else {
 
 #ifdef AF_INET6
-		sync_listen("::", conf.c_syncport, &sync_master6);
+		if (sync_master6.runs != SMS_RUNNING) {
+			sync_listen("::", conf.c_syncport, &sync_master6);
+			sync_master6_runs = sync_master6.runs;
+		}
 #endif
-		sync_listen("0.0.0.0", conf.c_syncport, &sync_master4);
+		if (sync_master4.runs != SMS_RUNNING) {
+			sync_listen("0.0.0.0", conf.c_syncport, &sync_master4);
+			sync_master4_runs = sync_master4.runs;
+		}
 	}
 
-
-	if (!sync_master4.runs && !sync_master6.runs) {
+	if (sync_master4.runs != SMS_RUNNING &&
+	    sync_master6.runs != SMS_RUNNING) {
 		mg_log(LOG_ERR, "cannot start MX sync, socket failed: %s",
 		    strerror(errno));
 		exit(EX_OSERR);
 	}
-	if (sync_master6.runs) {
+	if (sync_master6_runs == SMS_RUNNING) {
 		if ((error = pthread_create(&tid, NULL, sync_master,
-		    (void *)&sync_master6)) != 0) {
+		    (void *)&sync_master6)) != SMS_NOTRUNNING) {
 			mg_log(LOG_ERR, 
 			    "Cannot run MX sync thread for IPv6: %s",
 			    strerror(error));
@@ -647,7 +724,7 @@ sync_master_restart(void) {
 			exit(EX_OSERR);
 		}
 	}
-	if (sync_master4.runs) {
+	if (sync_master4_runs == SMS_RUNNING) {
 		if ((error = pthread_create(&tid, NULL, sync_master,
 		    (void *)&sync_master4)) != 0) {
 			mg_log(LOG_ERR, 
@@ -662,6 +739,8 @@ sync_master_restart(void) {
 			exit(EX_OSERR);
 		}
 	}
+last:
+	pthread_mutex_unlock(&sync_master_lock);
 }
 
 void *
@@ -670,6 +749,7 @@ sync_master(arg)
 {
 	struct sync_master_sock *sms = arg;
 
+	conf_retain();
 	for (;;) {
 		sockaddr_t raddr;
 		socklen_t raddrlen;
@@ -679,27 +759,41 @@ sync_master(arg)
 		struct peer *peer;
 		char peerstr[IPADDRSTRLEN];
 		int error;
+		int sock;
 
+		pthread_mutex_lock(&sync_master_lock);
+		sock = sms->sock;
+		pthread_mutex_unlock(&sync_master_lock);
+
+		/* TODO: accept connections in nonblocking mode
+		 * in order to watch conf change */
 		bzero((void *)&raddr, sizeof(raddr));
 		raddrlen = sizeof(raddr);
-		if ((fd = accept(sms->sock, SA(&raddr), &raddrlen)) == -1) {
-			mg_log(LOG_ERR, "incoming connexion "
+		if ((fd = accept(sock, SA(&raddr), &raddrlen)) == -1) {
+			mg_log(LOG_ERR, "incoming connection "
 			    "failed: %s", strerror(errno));
 
-			if (errno != ECONNABORTED)
-				exit(EX_OSERR);
-			continue;
+			if (is_fatal(errno))
+                        	exit(EX_OSERR);
+
+
 		}
 		unmappedaddr(SA(&raddr), &raddrlen);
 
+		conf_release();
+		conf_retain();
+
 		iptostring(SA(&raddr), raddrlen, peerstr, sizeof(peerstr));
-		mg_log(LOG_INFO, "Incoming MX sync connexion from %s", 
+		mg_log(LOG_INFO, "Incoming MX sync connection from %s", 
 		    peerstr);
 
-		if ((stream = fdopen(fd, "w+")) == NULL) {
+		errno = 0;
+		if ((stream = Fdopen(fd, "w+")) == NULL) {
 			mg_log(LOG_ERR, 
-			    "incoming connexion from %s failed, "
-			    "fdopen fail: %s", peerstr, strerror(errno));
+			    "incoming connection from %s failed, "
+			    "fdopen fail: %s", peerstr, 
+		    	    (errno == 0) ? "out of stdio streams"
+					 : strerror(errno));
 			close(fd);
 			exit(EX_OSERR);
 		}
@@ -717,10 +811,13 @@ sync_master(arg)
 			fprintf(stream, "105 No more peers, shutting down!\n");
 
 			PEER_UNLOCK;
-			fclose(stream);
+			Fclose(stream);
+			pthread_mutex_lock(&sync_master_lock);
 			close(sms->sock);
 			sms->sock = -1;
-			sms->runs = 0;
+			sms->runs = SMS_NOTRUNNING;
+			pthread_mutex_unlock(&sync_master_lock);
+			conf_release();
 			return NULL;
 		}
 			
@@ -772,20 +869,20 @@ sync_master(arg)
 			    peerstr);
 			fprintf(stream, 
 			    "106 You have no permission to talk, go away!\n");
-			fclose(stream);
+			Fclose(stream);
 			continue;
 		}
 
 		if ((error = pthread_create(&tid, NULL, 
 		    (void *(*)(void *))sync_server, (void *)stream)) != 0) {
-			mg_log(LOG_ERR, "incoming connexion from %s failed, "
+			mg_log(LOG_ERR, "incoming connection from %s failed, "
 			    "pthread_create failed: %s", 
 			    peerstr, strerror(error));
-			fclose(stream);
+			Fclose(stream);
 			continue;
 		}
 		if ((error = pthread_detach(tid)) != 0) {
-			mg_log(LOG_ERR, "incoming connexion from %s failed, "
+			mg_log(LOG_ERR, "incoming connection from %s failed, "
 			    "pthread_detach failed: %s",
 			    peerstr, strerror(error));
 			exit(EX_OSERR);
@@ -797,6 +894,7 @@ sync_master(arg)
 	return NULL;
 }
 
+/* sync_master_lock must be locked */
 static void
 sync_listen(addr, port, sms)
         char *addr, *port;
@@ -811,10 +909,10 @@ sync_listen(addr, port, sms)
 	int optval;
 	int s;
 
-	sms->runs = 1;
+	sms->runs = SMS_RUNNING;
 	laddrlen = sizeof(laddr);
 	if (ipfromstring(addr, SA(&laddr), &laddrlen, AF_UNSPEC) != 1) {
-		sms->runs = 0;
+		sms->runs = SMS_DISABLED;
 		return;
 	}
 
@@ -842,12 +940,12 @@ sync_listen(addr, port, sms)
 		break;
 #endif
 	default:
-		sms->runs = 0;
+		sms->runs = SMS_DISABLED;
 		return;
 	}
 
 	if ((s = socket(SA(&laddr)->sa_family, SOCK_STREAM, proto)) == -1) {
-		sms->runs = 0;
+		sms->runs = SMS_DISABLED;
 		return;
 	}
 
@@ -879,7 +977,7 @@ sync_listen(addr, port, sms)
 	if (bind(s, SA(&laddr), laddrlen) != 0) {
 		mg_log(LOG_ERR, "cannot start MX sync, bind failed: %s",
 		    strerror(errno));
-		sms->runs = 0;
+		sms->runs = SMS_DISABLED;
 		close(s);
 		return;
 	}
@@ -887,7 +985,7 @@ sync_listen(addr, port, sms)
 	if (listen(s, MXGLSYNC_BACKLOG) != 0) {
 		mg_log(LOG_ERR, "cannot start MX sync, listen failed: %s",
 		    strerror(errno));
-		sms->runs = 0;
+		sms->runs = SMS_DISABLED;
 		close(s);
 		return;
 	}
@@ -917,8 +1015,11 @@ sync_server(arg)
 	sockaddr_t addr;
 	socklen_t addrlen;
 	time_t date;
-	time_t aw = time(NULL) + conf.c_autowhite_validity;
+	time_t aw;
+	
+	conf_retain();
 
+	aw = time(NULL) + conf.c_autowhite_validity;
 	fprintf(stream, "200 Yeah, what do you want?\n");
 	fflush(stream);
 
@@ -1119,19 +1220,20 @@ eol:
 		fflush(stream);
 
 		if (action == PS_CREATE) {
-			PENDING_WRLOCK;
+			int dirty = 0;
+			PENDING_LOCK;
 			/* delay = -1 means unused: we supply the date */
-			pending_get(SA(&addr), addrlen, from, rcpt, date);
+			if (pending_get(SA(&addr), addrlen, from, rcpt, 
+					date, T_PENDING))
+				++dirty;
 			PENDING_UNLOCK;
+			dump_touch(dirty);
 		}
 		if (action == PS_DELETE) {
-			pending_del(SA(&addr), addrlen, from, rcpt, date);
-			autowhite_add(SA(&addr), addrlen, from, 
-			    rcpt, &aw, "(mxsync)");
+			pending_del(SA(&addr), addrlen, from, rcpt, date, aw);
 		}
 		if (action == PS_FLUSH) {
 			pending_del_addr(SA(&addr), addrlen, NULL, 0);
-			autowhite_del_addr(SA(&addr), addrlen);
 		}
 
 		/* Flush modifications to disk */
@@ -1139,7 +1241,9 @@ eol:
 	}
 
 	fprintf(stream, "202 Good bye\n");
-	fclose(stream);
+	Fclose(stream);
+
+	conf_release();
 
 	return;
 
@@ -1169,7 +1273,7 @@ sync_help(stream)
 	fprintf(stream, "203 \n");
 	fprintf(stream, "203 Available commands are:\n");
 	fprintf(stream, "203 help  -- displays this message\n");
-	fprintf(stream, "203 quit  -- terminate connexion\n");
+	fprintf(stream, "203 quit  -- terminate connection\n");
 	fprintf(stream, "203 vers2 -- speak version 2 protocol\n");
 	fprintf(stream, "203 vers3 -- speak version 3 protocol\n");
 	fprintf(stream, 
@@ -1190,10 +1294,10 @@ sync_help(stream)
 	return;
 }
 
-#define COM_TIMEOUT 3
 int
-sync_waitdata(fd)
+sync_waitdata(fd, stimeout)
 	int fd;
+	time_t stimeout;
 {
 	fd_set fdr, fde;
 	struct timeval timeout;
@@ -1205,10 +1309,23 @@ sync_waitdata(fd)
 	FD_ZERO(&fde);
 	FD_SET(fd, &fde);
 
-	timeout.tv_sec = COM_TIMEOUT;
+	timeout.tv_sec = stimeout;
 	timeout.tv_usec = 0;
 
-	retval = select(fd + 1, &fdr, NULL, &fde, &timeout); 
+	retval = select(fd + 1, &fdr, NULL, &fde, stimeout == 0 ? NULL : &timeout); 
+
+	if (retval == 0) {
+		mg_log(LOG_ERR, "sync_waitdata : select timeout");
+		
+	} else {
+		if (retval < 0) {
+			mg_log(LOG_ERR, "sync_waitdata : error");
+		} else {
+			if (FD_ISSET(fd,&fde)) {
+				mg_log(LOG_ERR, "sync_waitdata : exception fd is set");
+			}
+		}
+	}
 
 	return retval;
 }
@@ -1248,6 +1365,9 @@ sync_queue(peer, type, pending, autowhite)/* peer list must be read-locked */
 		sync_free(sync);
 	}
 
+	pthread_mutex_lock(&sync_dirty_lock);
+	sync_dirty = 1;
+	pthread_mutex_unlock(&sync_dirty_lock);
 	if ((error = pthread_cond_signal(&sync_sleepflag)) != 0) {
 		mg_log(LOG_ERR, 
 		    "cannot wakeup sync_sender: %s", strerror(error));
@@ -1306,9 +1426,14 @@ sync_sender(dontcare)
 	}
 
 	for (;;) {
-		if ((error = pthread_cond_wait(&sync_sleepflag, &mutex)) != 0)
-			mg_log(LOG_ERR, "pthread_cond_wait failed: %s", 
-			    strerror(error));
+		pthread_mutex_lock(&sync_dirty_lock);
+		while (!sync_dirty)
+			pthread_cond_wait(&sync_sleepflag, &sync_dirty_lock);
+		sync_dirty = 0;
+		pthread_mutex_unlock(&sync_dirty_lock);
+
+		conf_retain();
+
 		if (conf.c_debug) {
 			mg_log(LOG_DEBUG, "sync_sender running");
 			gettimeofday(&tv1, NULL);
@@ -1349,6 +1474,8 @@ out:
 			    "done %d entries in %ld.%06lds", done,
 			    tv3.tv_sec, tv3.tv_usec);
 		}
+
+		conf_release();
 	}
 }
 
@@ -1386,10 +1513,14 @@ local_addr(sa, salen)
 		return -1;
 	}
 
+	errno = 0;	 /* Solaris' bind() does not set errno... */
 	if (bind(sfd, sa, salen) == -1) {
 		if (errno != EADDRNOTAVAIL && 
 #ifdef __FreeBSD__
 		    errno != EINVAL &&
+#endif
+#ifdef __sun
+		    errno != 0 &&
 #endif
 		    1) {
 			mg_log(LOG_ERR, "local_addr: bind failed: %s",
@@ -1425,9 +1556,9 @@ select_protocol(peer, s, stream)
 
 		fflush(stream);
 
-		sync_waitdata(s);	
+		sync_waitdata(s, peer->p_socket_timeout);	
 		if (fgets(line, LINELEN, stream) == NULL) {
-			mg_log(LOG_ERR, "Lost connexion with peer %s: "
+			mg_log(LOG_ERR, "Lost connection with peer %s: "
 			    "%s (%d entries queued)", 
 			    peer->p_name, strerror(errno), peer->p_qlen);
 			return 0;
@@ -1435,7 +1566,7 @@ select_protocol(peer, s, stream)
 
 		if ((replystr = strtok_r(line, sep, &cookie)) == NULL) {
 			mg_log(LOG_ERR, "Unexpected reply \"%s\" from peer %s "
-			    "closing connexion (%d entries queued)", 
+			    "closing connection (%d entries queued)", 
 			    line, peer->p_name, peer->p_qlen);
 			return 0;
 		}
@@ -1477,3 +1608,15 @@ out:
 	
 	return;
 }
+
+static int 
+is_fatal(err)
+	int err;
+{
+	int *i;
+
+	for (i = sync_ignored_errno; *i; i++)
+		if (errno == *i)
+			return 0;
+	return 1;
+}
